import pickle

import pandas as pd
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import OneHotEncoder

from player_performance_ratings.cross_validator.cross_validator import MatchKFoldCrossValidator
from player_performance_ratings.pipelinefactory import Pipeline
from player_performance_ratings.transformation.pre_transformers import ConvertDataFrameToCategoricalTransformer, \
    SkLearnTransformerWrapper

from sklearn.metrics import log_loss

from player_performance_ratings.predictor import GameTeamPredictor
from player_performance_ratings.scorer.score import SklearnScorer

from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator
from player_performance_ratings.ratings.rating_calculators.performance_predictor import \
    RatingDifferencePerformancePredictor

from player_performance_ratings.transformation import LagTransformer, RollingMeanTransformer

from player_performance_ratings.ratings import ColumnWeight, UpdateRatingGenerator
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.rating_generator_tuner import UpdateRatingGeneratorTuner

from player_performance_ratings import ColumnNames
from player_performance_ratings.tuner.utils import ParameterSearchRange, get_default_team_rating_search_range, \
    get_default_lgbm_classifier_search_range_by_learning_rate
from venn_abers import VennAbersCalibrator

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.predictor.sklearn_models import SkLearnWrapper

from player_performance_ratings.tuner import PipelineTuner

column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_id",
    performance="plus_minus_per_minute",
    participation_weight="participation_weight",
    projected_participation_weight="projected_participation_weight",
    team_players_playing_time="team_player_mins",
    opponent_players_playing_time="opp_player_mins",
)

df = pd.read_pickle("data/game_player_full_with_minutes_prediction.pickle")

df[PredictColumnNames.TARGET] = df['won']
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])
df = df[df['game_minutes'] > 46]
df = df.rename(columns={'__target_prediction': 'minutes_prediction'})
df['plus_minus_per_minute'] = df['plus_minus'] / df['game_minutes']
df['participation_weight'] = df['minutes'] / df['game_minutes']
df['projected_participation_weight'] = df['minutes_prediction'] / df['game_minutes'].mean()
df['score_difference'] = df['score'] - df['score_opponent']

df = (
    df.assign(team_count=df.groupby('game_id')['team_id'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)
df = df.drop_duplicates(subset=['game_id', 'player_id'])


df.loc[df['minutes'] > 0, 'plus_minus_per_minute'] = df['plus_minus'] / df['minutes']
df.loc[df['minutes'] == 0, 'plus_minus_per_minute'] = 0

estimator = SkLearnWrapper(
    VennAbersCalibrator(
        estimator=LGBMClassifier(max_depth=2, learning_rate=0.1, n_estimators=55, verbose=-100, reg_alpha=1),
        inductive=True, cal_size=0.2, random_state=101))

# estimator = LGBMClassifier(max_depth=2, learning_rate=0.1, n_estimators=35, verbose=-100, reg_alpha=1)

estimator = LogisticRegression()

predictor = GameTeamPredictor(
    estimator=estimator,
    game_id_colum="game_id",
    team_id_column="team_id",
   # categorical_transformers=[ConvertDataFrameToCategoricalTransformer(features=["location"])],
    categorical_transformers=[SkLearnTransformerWrapper(transformer=OneHotEncoder(), features=["location"])]
)

performance_predictor = RatingDifferencePerformancePredictor(
    rating_diff_team_from_entity_coef=0.0055,
)

rating_generator = UpdateRatingGenerator(column_names=column_names, match_rating_generator=MatchRatingGenerator(
    performance_predictor=performance_predictor))

column_weights = [ColumnWeight(name='plus_minus', weight=1)]

post_rating_transformers = [
    LagTransformer(
        features=["score_difference"],
        lag_length=5,
        granularity=[column_names.player_id],
        column_names=column_names,
        days_between_lags=[1, 2, 3, 4, 5],
    ),
    LagTransformer(
        features=[],
        lag_length=3,
        granularity=[column_names.player_id],
        column_names=column_names,
        prefix="future_lag",
        future_lag=True,
        days_between_lags=[1, 2, 3]
    ),
]

pipeline = Pipeline(
    post_rating_transformers=post_rating_transformers,
    rating_generators=rating_generator,
    column_weights=column_weights,
    predictor=predictor,
)

rating_generator_tuner = UpdateRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
   # start_rating_search_ranges=start_rating_search_range,
    team_rating_n_trials=35,
    start_rating_n_trials=10,
)
predictor_tuner = PredictorTuner(
    default_params={'learning_rate': 0.07},
    search_ranges=get_default_lgbm_classifier_search_range_by_learning_rate(learning_rate=0.07),
    n_trials=30,
    date_column_name=column_names.start_date,
)

cross_validator = MatchKFoldCrossValidator(
    scorer=SklearnScorer(pred_column=pipeline._predictor.pred_column, scorer_function=log_loss),
    match_id_column_name=column_names.match_id,
    n_splits=5,
    date_column_name=column_names.start_date
)

tuner = PipelineTuner(
    pipeline=pipeline,
    fit_best=True,
    cross_validator=cross_validator,
    rating_generator_tuners=rating_generator_tuner,
    predictor_tuner=predictor_tuner,
)
best_match_predictor = tuner.tune(df=df)
game_player_predictions = best_match_predictor.train(df=df)

pickle.dump(best_match_predictor, open("models/nba_game_winner", 'wb'))
game_player_predictions.to_pickle("data/game_player_predictions.pickle")

game_player_cv = best_match_predictor.generate_cross_validate_df(df=df, cross_validator=cross_validator)
game_player_cv.to_pickle("data/game_player_cv.pickle")



from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.enums import RatingEstimatorFeatures, RatingHistoricalFeatures

from player_performance_ratings.data_structures import Match, PlayerRating, \
    TeamRating, ColumnNames


class RatingGenerator(ABC):

    def __init__(self, column_names: ColumnNames):
        self.column_names = column_names

    @abstractmethod
    def generate_historical(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None) -> dict[
        RatingEstimatorFeatures, list[float]]:
        pass

    @abstractmethod
    def generate_future(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None) -> dict[
        RatingEstimatorFeatures, list[float]]:
        pass

    @property
    @abstractmethod
    def player_ratings(self) -> dict[str, PlayerRating]:
        pass

    @property
    @abstractmethod
    def team_ratings(self) -> list[TeamRating]:
        pass

    @property
    @abstractmethod
    def estimator_features_out(self) -> list[RatingEstimatorFeatures]:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[Union[RatingEstimatorFeatures, RatingHistoricalFeatures]]:
        pass






import copy
from typing import Optional, List, Union



from player_performance_ratings import Pipeline
from player_performance_ratings.ratings import PerformancesGenerator, ColumnWeight

from player_performance_ratings.predictor import BaseMLWrapper
from player_performance_ratings.ratings.rating_generator import RatingGenerator
from player_performance_ratings.transformation.base_transformer import BaseTransformer, BasePostTransformer


class PipelineFactory():

    def __init__(self,
                 predictor: BaseMLWrapper,
                 rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]] = None,
                 performances_generator: Optional[PerformancesGenerator] = None,
                 post_rating_transformers: Optional[List[BasePostTransformer]] = None,
                 ):

        self.post_rating_transformers = post_rating_transformers or []


        self.rating_generators = rating_generators or []
        if isinstance(self.rating_generators, RatingGenerator):
            self.rating_generators = [self.rating_generators]

        self.predictor = predictor
        self.performances_generator = performances_generator


    def create(self,
               performances_generator: Optional[PerformancesGenerator] = None,
               rating_generators: Optional[list[RatingGenerator]] = None,
               post_rating_transformers: Optional[List[BaseTransformer]] = None,
               predictor: Optional[BaseMLWrapper] = None,
               ) -> Pipeline:

        rating_generators = rating_generators if rating_generators is not None else [copy.deepcopy(r) for r in self.rating_generators]
        performances_generator = performances_generator if performances_generator is not None else copy.deepcopy(self.performances_generator)
        post_rating_transformers = post_rating_transformers if post_rating_transformers is not None else [copy.deepcopy(r) for r in self.post_rating_transformers]
        predictor = predictor if predictor is not None else self.predictor

        return Pipeline(
            rating_generators=rating_generators,
            performances_generator=performances_generator,
            post_rating_transformers=copy.deepcopy(post_rating_transformers),
            predictor=predictor,
        )



from abc import abstractmethod, ABC

import pandas as pd
from player_performance_ratings.transformation.base_transformer import BaseTransformer

from player_performance_ratings.predictor import BaseMLWrapper

from player_performance_ratings.scorer import BaseScorer


class CrossValidator(ABC):

    def __init__(self, scorer: BaseScorer):
        self.scorer = scorer


    @abstractmethod
    def generate_validation_df(self,
                               df: pd.DataFrame,
                               post_transformers: list[BaseTransformer],
                               predictor: BaseMLWrapper,
                               estimator_features: list[str],
                               ) -> pd.DataFrame:
        pass


    def cross_validation_score(self, validation_df: pd.DataFrame) -> float:
        return self.scorer.score(df=validation_df)




import pandas as pd
from lightgbm import LGBMClassifier

from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.utils import get_default_team_rating_search_range, ParameterSearchRange, \
    get_default_lgbm_classifier_search_range

from player_performance_ratings.predictor.transformer import ConvertDataFrameToCategoricalTransformer
from player_performance_ratings.ratings.rating_calculators.performance_predictor import \
    RatingNonOpponentPerformancePredictor
from player_performance_ratings.scorer import OrdinalLossScorer

from player_performance_ratings import ColumnNames, Pipeline
from player_performance_ratings.cross_validator.cross_validator import MatchKFoldCrossValidator
from player_performance_ratings.predictor import Predictor, SkLearnWrapper
from player_performance_ratings.ratings import ColumnWeight, UpdateRatingGenerator, PerformancesGenerator, \
    RatingEstimatorFeatures
from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator, RatingMeanPerformancePredictor, \
    StartRatingGenerator
from player_performance_ratings.tuner import PipelineTuner, PerformancesGeneratorTuner
from player_performance_ratings.tuner.rating_generator_tuner import UpdateRatingGeneratorTuner

df = pd.read_pickle(
    r"C:\Users\Admin\PycharmProjects\rating-model-transformer\examples\lol\data\lol_game_winner_total_kills_df")

df = df[df['date'] > '2021-01-01']
df['league_position'] = df['league'] + '__' + df['position']
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])



df = (
    df.loc[lambda x: x.position != 'team']
    .assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)


column_names_weighted = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='performance',
    league='league_position',
    position='position',
)

column_names_kpm = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='performance_kills',
    league='league_position',
    position='position',
)


df = df[
    [column_names_weighted.team_id, column_names_weighted.match_id, column_names_weighted.start_date, column_names_weighted.player_id,
     "kills", "deaths",
     "damagetochampions", "result", "position", "league", "champion", "gamelength", "assists", "duration",
     "kills_rating_avg", "prob",'league_position',
     "total_kills"]]

df['__target'] = df['kills']
df["__target"] = df['__target'].clip(0, 9)

df['kills_per_minute'] = df['kills'] / df['gamelength'] * 60

estimator = LGBMClassifier(max_depth=2, n_estimators=60, verbose=-100)

predictor = Predictor(
    estimator=estimator,
    estimator_features=[
        "prob",
        "kills_rating_avg",
        "position"
    ],
    categorical_transformers=[ConvertDataFrameToCategoricalTransformer(features=["position"])],
)



rating_generator_weighted = UpdateRatingGenerator(
    column_names=column_names_weighted,
    match_rating_generator=MatchRatingGenerator(
        league_rating_adjustor_multiplier=0,
        performance_predictor=RatingNonOpponentPerformancePredictor(),
        start_rating_generator=StartRatingGenerator(league_quantile=0.5, min_count_for_percentiles=5),
    ),
)

rating_generator_kpm = UpdateRatingGenerator(
    column_names=column_names_kpm ,
    match_rating_generator=MatchRatingGenerator(
        league_rating_adjustor_multiplier=0,
        performance_predictor=RatingNonOpponentPerformancePredictor(),
        start_rating_generator=StartRatingGenerator(league_quantile=0.5, min_count_for_percentiles=5),
    ),
)
column_weights = [
    [ColumnWeight(name='kills_per_minute', weight=1), ColumnWeight(name='kills', weight=1), ColumnWeight(name='assists', weight=1)],
    [ColumnWeight("kills", weight=1)],
]

pipeline = Pipeline(
    rating_generators=[rating_generator_weighted, rating_generator_kpm],
    predictor=predictor,
    performances_generator=PerformancesGenerator(
        column_names=[column_names_weighted, column_names_kpm],
        column_weights=column_weights
    )
)


weighter_search_range = {
    column_names_weighted.performance:
        [
            ParameterSearchRange(
                name='kills_per_minute',
                type='uniform',
                low=0,
                high=1
            ),
            ParameterSearchRange(
                name='kills',
                type='uniform',
                low=0,
                high=1
            ),
            ParameterSearchRange(
                name='assists',
                type='uniform',
                low=0,
                high=0.1
            ),
        ]
}

performance_generator_tuner = PerformancesGeneratorTuner(performances_weight_search_ranges=weighter_search_range,
                                                         n_trials=15)

cross_validator = MatchKFoldCrossValidator(
    scorer=OrdinalLossScorer(pred_column=pipeline.predictor.pred_column, targets_to_measure=[t for t in range(9)]),
    match_id_column_name=column_names_weighted.match_id,
    n_splits=1,
    date_column_name=column_names_weighted.start_date,
)

rating_generator_tuner_weighted = UpdateRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
    team_rating_n_trials=20,
    start_rating_n_trials=0,
)
predictor_tuner = PredictorTuner(
    search_ranges=get_default_lgbm_classifier_search_range(),
    date_column_name=column_names_weighted.start_date,
)
tuner = PipelineTuner(
    performances_generator_tuners=[performance_generator_tuner, None],
    rating_generator_tuners=[rating_generator_tuner_weighted, None],
    fit_best=True,
    pipeline=pipeline,
    predictor_tuner=predictor_tuner,
    cross_validator=cross_validator,
)
tuner.tune(df)



import copy

from typing import Optional, Union

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial

from player_performance_ratings.cross_validator.cross_validator import CrossValidator
from player_performance_ratings.ratings import ColumnWeight, PerformancesGenerator
from player_performance_ratings.ratings.enums import RatingEstimatorFeatures
from player_performance_ratings.transformation.base_transformer import BaseTransformer

from player_performance_ratings import PipelineFactory

from player_performance_ratings.tuner.utils import add_params_from_search_range, ParameterSearchRange

RC = RatingEstimatorFeatures


class PerformancesGeneratorTuner:

    def __init__(self,
                 performances_weight_search_ranges: dict[str, list[ParameterSearchRange]],
                 feature_names: Optional[Union[list[str], list[list[str]]]] = None,
                 lower_is_better_features: Optional[list[str]] = None,
                 n_trials: int = 30,
                 ):

        self.performances_weight_search_ranges = performances_weight_search_ranges
        self.feature_names = feature_names
        self.lower_is_better_features = lower_is_better_features

        self.n_trials = n_trials
        self._scores = []

    def tune(self,
             df: pd.DataFrame,
             rating_idx: int,
             cross_validator: CrossValidator,
             pipeline_factory: PipelineFactory,
             ) -> PerformancesGenerator:

        if pipeline_factory.performances_generator is None:
            raise ValueError("pipeline_factory.performances_generator is None. Please provide a performances_generator.")

        df = df.copy()

        column_names = [r.column_names for r in pipeline_factory.rating_generators]

        pipeline_factory = copy.deepcopy(pipeline_factory)

        def objective(trial: BaseTrial,
                      df: pd.DataFrame,
                      pipeline_factory: PipelineFactory,
                      ) -> float:

            best_pre_transformers = copy.deepcopy(pipeline_factory.performances_generator.pre_transformations)
            column_weights = [cw for cw in pipeline_factory.performances_generator.column_weights]
            for performance_name, search_range in self.performances_weight_search_ranges.items():
                column_weights[rating_idx] = []
                raw_params = {}
                renamed_search = copy.deepcopy(search_range)
                for sr in renamed_search:
                    sr.name = f"{performance_name}__{sr.name}"
                add_params_from_search_range(trial=trial, params=raw_params, parameter_search_range=renamed_search)
                new_col_weights = self._create_column_weights(params=raw_params, remove_string=f"{performance_name}__")
                for new_col_weight in new_col_weights:
                    if new_col_weight.name in [cw.name for cw in column_weights[rating_idx]]:
                        column_weights[rating_idx][[cw.name for cw in column_weights[rating_idx]].index(new_col_weight.name)] = new_col_weight
                    else:
                        column_weights[rating_idx].append(new_col_weight)

                col_names = [r.column_names for r in pipeline_factory.rating_generators]

                performances_generator = PerformancesGenerator(
                column_names = col_names,
                column_weights = column_weights,
                pre_transformations = best_pre_transformers,
            )
            pipeline = pipeline_factory.create(performances_generator=performances_generator)
            return pipeline.cross_validate_score(df=df, cross_validator=cross_validator,
                                                 create_performance=True, create_rating_features=True)

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df=df,
                                    pipeline_factory=pipeline_factory),
            n_trials=self.n_trials, callbacks=callbacks)

        best_params = study.best_params
        best_column_weights = self._select_best_column_weights(all_params=best_params)

        return PerformancesGenerator(column_weights=best_column_weights,
                                     column_names=column_names,
                                     pre_transformations=pipeline_factory.performances_generator.pre_transformations,
                                     )

    def _create_column_weights(self, params: dict, remove_string: str) -> list[ColumnWeight]:

        sum_weights = sum([v for _, v in params.items()])
        column_weights = []
        for name, weight in params.items():
            column_weights.append(ColumnWeight(name=name.replace(remove_string, ""), weight=weight / sum_weights))

        return column_weights

    def _select_best_column_weights(self, all_params: dict) -> list[list[ColumnWeight]]:
        best_column_weights = []
        for performance_name in self.performances_weight_search_ranges:
            column_weights = []
            sum_weights = 0
            for param in all_params:

                if f"{performance_name}__" in param:
                    sum_weights += all_params[param]

            for param in all_params:
                if f"{performance_name}__" in param:
                    column_weights.append(ColumnWeight(name=param.replace(f"{performance_name}__", ""),
                                                       weight=all_params[param] / sum_weights))

            best_column_weights.append(column_weights)

        return best_column_weights



import copy
from typing import Optional

import pandas as pd
from player_performance_ratings.transformation.base_transformer import BaseTransformer

from player_performance_ratings.cross_validator._base import CrossValidator
from player_performance_ratings.predictor import BaseMLWrapper
from player_performance_ratings.scorer import BaseScorer


class MatchCountCrossValidator(CrossValidator):

    def __init__(self,
                 scorer: BaseScorer,
                 match_id_column_name: str,
                 validation_match_count: int,
                 n_splits: int = 3):
        super().__init__(scorer=scorer)
        self.n_splits = n_splits
        self.match_id_column_name = match_id_column_name
        self.validation_match_count = validation_match_count

    def generate_validation_df(self,
                               df: pd.DataFrame,
                               post_transformers: list[BaseTransformer],
                               predictor: BaseMLWrapper,
                               estimator_features: list[str]) -> pd.DataFrame:
        validation_dfs = []
        df = df.assign(__cv_match_number=pd.factorize(df[self.match_id_column_name])[0])
        max_match_number = df['__cv_match_number'].max()
        train_cut_off_match_number = max_match_number - self.validation_match_count * self.n_splits + 1
        step_matches = self.validation_match_count

        train_df = df[(df['__cv_match_number'] < train_cut_off_match_number)]
        if len(train_df) < 0:
            raise ValueError(
                f"train_df is empty. train_cut_off_day_number: {train_cut_off_match_number}. Select a lower validation_match value.")
        validation_df = df[(df['__cv_match_number'] >= train_cut_off_match_number) & (
                df['__cv_match_number'] < train_cut_off_match_number + step_matches)]

        for _ in range(self.n_splits):
            for post_transformer in post_transformers:
                train_df = post_transformer.fit_transform(train_df)
                validation_df = post_transformer.transform(validation_df)
            predictor.train(train_df, estimator_features=estimator_features)
            validation_df = predictor.add_prediction(validation_df)
            validation_dfs.append(validation_df)

            train_cut_off_match_number = train_cut_off_match_number + step_matches
            train_df = df[(df['__cv_match_number'] < train_cut_off_match_number)]
            validation_df = df[(df['__cv_match_number'] >= train_cut_off_match_number) & (
                    df['__cv_match_number'] < train_cut_off_match_number + step_matches)]

        return pd.concat(validation_dfs)


class MatchKFoldCrossValidator(CrossValidator):
    def __init__(self,
                 scorer: BaseScorer,
                 match_id_column_name: str,
                 date_column_name: str,
                 min_validation_date: Optional[str] = None,
                 n_splits: int = 3):
        super().__init__(scorer=scorer)
        self.match_id_column_name = match_id_column_name
        self.date_column_name = date_column_name
        self.n_splits = n_splits
        self.min_validation_date = min_validation_date

    def generate_validation_df(self,
                               df: pd.DataFrame,
                               predictor: BaseMLWrapper,
                               estimator_features: Optional[list[str]] = None,
                               post_transformers: Optional[list[BaseTransformer]] = None,
                               ) -> pd.DataFrame:
        validation_dfs = []

        estimator_features = estimator_features or []
        post_transformers = post_transformers or []

        if not self.min_validation_date:
            unique_dates = df[self.date_column_name].unique()
            median_number = len(unique_dates) // 2
            self.min_validation_date = unique_dates[median_number]

        df = df.assign(__cv_match_number=range(len(df)))
        min_validation_match_number = df[df[self.date_column_name] >= self.min_validation_date][
            "__cv_match_number"].min()

        max_match_number = df['__cv_match_number'].max()
        train_cut_off_match_number = min_validation_match_number
        step_matches = (max_match_number - min_validation_match_number) / self.n_splits
        train_df = df[(df['__cv_match_number'] < train_cut_off_match_number)]
        if len(train_df) < 0:
            raise ValueError(
                f"train_df is empty. train_cut_off_day_number: {train_cut_off_match_number}. Select a lower validation_match value.")
        validation_df = df[(df['__cv_match_number'] >= train_cut_off_match_number) & (
                df['__cv_match_number'] < train_cut_off_match_number + step_matches)]

        for idx in range(self.n_splits):

            for post_transformer in post_transformers:
                if hasattr(post_transformer, "_df"):
                    post_transformer._df = None
                train_df = post_transformer.fit_transform(train_df)
                validation_df = post_transformer.transform(validation_df)

            predictor.train(train_df, estimator_features=estimator_features)
            validation_df = predictor.add_prediction(validation_df)
            validation_dfs.append(validation_df)

            train_cut_off_match_number = train_cut_off_match_number + step_matches
            train_df = df[(df['__cv_match_number'] < train_cut_off_match_number)]

            if idx == self.n_splits - 2:
                validation_df = df[(df['__cv_match_number'] >= train_cut_off_match_number)]
            else:
                validation_df = df[(df['__cv_match_number'] >= train_cut_off_match_number) & (
                        df['__cv_match_number'] < train_cut_off_match_number + step_matches)]

        return pd.concat(validation_dfs).drop(columns=['__cv_match_number'])



from dataclasses import dataclass
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Callable, Union, Any

import numpy as np
import pandas as pd

from player_performance_ratings.consts import PredictColumnNames


class Operator(Enum):
    EQUALS = '=='
    NOT_EQUALS = '!='
    GREATER_THAN = '>'
    LESS_THAN = '<'
    GREATER_THAN_OR_EQUALS = '>='
    LESS_THAN_OR_EQUALS = '<='
    IN = 'in'
    NOT_IN = 'not in'


@dataclass
class Filter:
    column_name: str
    value: Union[Any, list[Any]]
    operator: Operator


def apply_filters(df: pd.DataFrame, filters: list[Filter]) -> pd.DataFrame:
    df = df.copy()
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df[df[filter.column_name] == filter.value]
        elif filter.operator == Operator.NOT_EQUALS:
            df = df[df[filter.column_name] != filter.value]
        elif filter.operator == Operator.GREATER_THAN:
            df = df[df[filter.column_name] > filter.value]
        elif filter.operator == Operator.LESS_THAN:
            df = df[df[filter.column_name] < filter.value]
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df[df[filter.column_name] >= filter.value]
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df[df[filter.column_name] <= filter.value]
        elif filter.operator == Operator.IN:
            df = df[df[filter.column_name].isin(filter.value)]
        elif filter.operator == Operator.NOT_IN:
            df = df[~df[filter.column_name].isin(filter.value)]

    return df


class BaseScorer(ABC):

    def __init__(self, target: str, pred_column: str, filters: Optional[list[Filter]] = None,
                 granularity: Optional[list[str]] = None):
        self.target = target
        self.pred_column = pred_column
        self.filters = filters or []
        self.granularity = granularity

    @abstractmethod
    def score(self, df: pd.DataFrame) -> float:
        pass


class SklearnScorer(BaseScorer):

    def __init__(self,
                 pred_column: str,
                 scorer_function: Callable,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 granularity: Optional[list[str]] = None,
                 filters: Optional[list[Filter]] = None
                 ):
        self.pred_column_name = pred_column
        self.scorer_function = scorer_function
        super().__init__(target=target, pred_column=pred_column, granularity=granularity, filters=filters)

    def score(self, df: pd.DataFrame) -> float:
        df = df.copy()
        df = apply_filters(df, self.filters)
        if self.granularity:
            grouped = df.groupby(self.granularity)[self.pred_column_name, self.target].mean().reset_index()
        else:
            grouped = df
        if isinstance(df[self.pred_column_name].iloc[0], list):
            return self.scorer_function(grouped[self.target], np.asarray(grouped[self.pred_column_name]).tolist())
        return self.scorer_function(grouped[self.target], grouped[self.pred_column_name])


class OrdinalLossScorer(BaseScorer):

    def __init__(self,
                 pred_column: str,
                 targets_to_measure: Optional[list[int]] = None,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 granularity: Optional[list[str]] = None,
                 filters: Optional[list[Filter]] = None
                 ):

        self.pred_column_name = pred_column
        self.targets_to_measure = targets_to_measure
        self.granularity = granularity
        super().__init__(target=target, pred_column=pred_column, filters=filters, granularity=granularity)

    def score(self, df: pd.DataFrame) -> float:
        if self.targets_to_measure is None:
            self.targets_to_measure = df[self.target].unique().tolist()
        self.targets_to_measure.sort()
        df = df.copy()
        df = apply_filters(df, self.filters)
        df.reset_index(drop=True, inplace=True)
        probs = df[self.pred_column_name]
        last_column_name = f'prob_under_{self.targets_to_measure[0] - 0.5}'
        df[last_column_name] = probs.apply(lambda x: x[0])

        sum_lr = 0

        for idx, class_ in enumerate(self.targets_to_measure[1:]):

            p_c = 'prob_under_' + str(class_ + 0.5)
            df[p_c] = probs.apply(lambda x: x[idx+1]) + df[last_column_name]

            count_exact = len(df[df['__target'] == class_])
            weight_class = count_exact / len(df)

            if self.granularity:
                grouped = df.groupby(self.granularity + ['__target'])[p_c].mean().reset_index()
            else:
                grouped = df

            grouped['min'] = 0.0001
            grouped['max'] = 0.9999
            grouped[p_c] = np.minimum(grouped['max'], grouped[p_c])
            grouped[p_c] = np.maximum(grouped['min'], grouped[p_c])
            grouped['log_loss'] = 0
            grouped.loc[grouped['__target'] <= class_, 'log_loss'] = np.log(grouped[p_c])
            grouped.loc[grouped['__target'] > class_, 'log_loss'] = np.log(1 - grouped[p_c])
            log_loss = grouped['log_loss'].mean()
            sum_lr -= log_loss * weight_class

            last_column_name = p_c

        return sum_lr



import logging
from dataclasses import dataclass
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.transformation.pre_transformers import \
    SymmetricDistributionTransformer, MinMaxTransformer, PartialStandardScaler

PartialStandardScaler

from player_performance_ratings import ColumnNames

from player_performance_ratings.transformation.base_transformer import BaseTransformer


@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False

    def __post_init__(self):
        if self.weight < 0:
            raise ValueError("Weight must be positive")
        if self.weight > 1:
            raise ValueError("Weight must be less than 1")


def auto_create_pre_performance_transformations(
        pre_transformations: list[BaseTransformer],
        column_weights: Union[list[list[ColumnWeight]], list[ColumnWeight]],
        column_names: list[ColumnNames],
) -> list[BaseTransformer]:
    """
    Creates a list of transformers that ensure the performance column is generated in a way that makes sense for the rating model.
    Ensures columns aren't too skewed, scales them to similar ranges, ensure values are between 0 and 1,
    and then weights them according to the column_weights.
    """

    if not isinstance(column_weights[0], list):
        column_weights = [column_weights]


    contains_not_position = True if any([c.position is None for c in column_names]) else False

    not_transformed_features = []
    transformed_features = []


    for idx, col_weights in enumerate(column_weights):
        not_transformed_features += [c.name for c in column_weights[idx] if c.name not in not_transformed_features]


        if column_names[idx].position is not None:
            granularity = [column_names[idx].position]
            features_in = []
            for column_weight in col_weights:
                features_in.append(column_weight.name)
                not_transformed_features.remove(column_weight.name)

            distribution_transformer = SymmetricDistributionTransformer(
                features=features_in,
                granularity=granularity, prefix="")
            transformed_features += [distribution_transformer.prefix + f for f in distribution_transformer.features]
            pre_transformations.append(distribution_transformer)


    if contains_not_position:
        distribution_transformer = SymmetricDistributionTransformer(features=not_transformed_features, prefix="")
        pre_transformations.append(distribution_transformer)

    all_features = transformed_features + not_transformed_features

    pre_transformations.append(PartialStandardScaler(features=all_features, ratio=1, max_value=9999, target_mean=0, prefix=""))
    pre_transformations.append(MinMaxTransformer(features=all_features))

    return pre_transformations


class PerformancesGenerator():

    def __init__(self,
                 column_weights: Union[list[list[ColumnWeight]], list[ColumnWeight]],
                 column_names: Union[list[ColumnNames], ColumnNames],
                 pre_transformations: Optional[list[BaseTransformer]] = None,
                 auto_transform_performance: bool = True
                 ):
        self.column_names = column_names if isinstance(column_names, list) else [column_names]
        self.column_weights = column_weights if isinstance(column_weights[0], list) else [column_weights]
        if len(self.column_names) != len(self.column_weights):
            raise ValueError("column_names and column_weights must be the same length, and both equal to the number of rating generators")
        self.auto_transform_performance = auto_transform_performance

        self.pre_transformations = pre_transformations or []
        if self.auto_transform_performance:
            self.pre_transformations = auto_create_pre_performance_transformations(
                pre_transformations=self.pre_transformations, column_weights=column_weights,
                column_names=self.column_names)

    def generate(self, df):

        if self.pre_transformations:
            for pre_transformation in self.pre_transformations:
                df = pre_transformation.fit_transform(df)

        for idx, col_name in enumerate(self.column_names):
            if self.pre_transformations:
                max_idx = len(self.pre_transformations) - 1
                column_weighs_mapping = {col_weight.name: self.pre_transformations[max_idx].features_out[idx] for
                                         idx, col_weight in enumerate(self.column_weights[idx])}
            else:
                column_weighs_mapping = None

            df[col_name.performance] = self._weight_columns(df=df, col_name=col_name,
                                                            col_weights=self.column_weights[idx],
                                                            column_weighs_mapping=column_weighs_mapping)

            if df[col_name.performance].isnull().any():
                logging.error(
                    f"df[{col_name.performance}] contains nan values. Make sure all column_names used in column_weights are imputed beforehand")
                raise ValueError("performance contains nan values")

        return df

    def _weight_columns(self,
                        df: pd.DataFrame,
                        col_name: ColumnNames,
                        col_weights: list[ColumnWeight],
                        column_weighs_mapping: dict[str, str]
                        ) -> pd.DataFrame:
        df = df.copy()
        df[f"__{col_name.performance}"] = 0

        df['sum_cols_weights'] = 0
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f'weight__{column_weight.name}'] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df['sum_cols_weights'] = df['sum_cols_weights'] + df[f'weight__{column_weight.name}']

        drop_cols = ['sum_cols_weights', f"__{col_name.performance}"]
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] / df['sum_cols_weights']
            drop_cols.append(f'weight__{column_weight.name}')

        sum_weight = sum([w.weight for w in col_weights])

        for column_weight in col_weights:

            if column_weighs_mapping:
                feature_name = column_weighs_mapping[column_weight.name]
            else:
                feature_name = column_weight.name

            if column_weight.lower_is_better:
                df[f"__{col_name.performance}"] += df[f'weight__{column_weight.name}']/sum_weight * (
                        1 - df[feature_name])
            else:
                df[f"__{col_name.performance}"] += df[f'weight__{column_weight.name}']/sum_weight * df[feature_name]

        return df[f"__{col_name.performance}"]

    @property
    def features_out(self) -> list[str]:
        return [c.performance for c in self.column_names]



from typing import Optional, Any, Union

import numpy as np
import pandas as pd

from player_performance_ratings.ratings import convert_df_to_matches
from player_performance_ratings.ratings.rating_calculators import RatingMeanPerformancePredictor
from player_performance_ratings.ratings.rating_calculators.performance_predictor import \
    RatingNonOpponentPerformancePredictor
from player_performance_ratings.ratings.rating_calculators.match_rating_generator import MatchRatingGenerator
from player_performance_ratings.ratings.enums import RatingEstimatorFeatures, RatingHistoricalFeatures

from player_performance_ratings.data_structures import Match, PreMatchRating, PreMatchTeamRating, PlayerRating, \
    TeamRating, ColumnNames, TeamRatingChange
from player_performance_ratings.ratings.rating_generator import RatingGenerator


class UpdateRatingGenerator(RatingGenerator):
    """
    Generates ratings for players and teams based on the match-performance of the player and the ratings of the players and teams.
    Ratings are updated after a match is finished
    """

    def __init__(self,
                 column_names: ColumnNames,
                 match_rating_generator: MatchRatingGenerator = MatchRatingGenerator(),
                 estimator_features_out: Optional[list[RatingEstimatorFeatures]] = None,
                 historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
                 estimator_features_pass_through: Optional[list[RatingEstimatorFeatures]] = None,
                 distinct_positions: Optional[list[str]] = None,
                 ):

        """

        :param match_rating_generator:
            The class contains the logic for generating and updating team ratings and contains many parameters that can be tuned.
        :param features_names_created:
            If called by match_predictor, feature_names_created determines which features will be used for prediction.
            If other features such as player_rating_difference is used, it must be added to this list.
        """
        super().__init__(column_names=column_names)
        self.team_rating_generator = match_rating_generator
        self.distinct_positions = distinct_positions
        self._estimator_features_pass_through = estimator_features_pass_through or []

        self._estimator_features_out = estimator_features_out if estimator_features_out is not None else [
            RatingEstimatorFeatures.RATING_MEAN_PROJECTED] if isinstance(match_rating_generator.performance_predictor,
                                                                         RatingMeanPerformancePredictor) else [
            RatingEstimatorFeatures.PLAYER_RATING] if isinstance(match_rating_generator.performance_predictor,
                                                                 RatingNonOpponentPerformancePredictor) else [
            RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED]

        if self.distinct_positions:
            self._estimator_features_out += [RatingEstimatorFeatures.RATING_DIFFERENCE_POSITION + "_" + p for p in
                                             self.distinct_positions]

        self._historical_features_out = historical_features_out or []


        # If projected participation weight is not None, then the projected ratings will be used instead of the actual ratings (which first are known after game is finished)

        self.ratings_df = None

    def generate_historical(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None) -> dict[Union[
                                                                                                                        RatingEstimatorFeatures, RatingHistoricalFeatures], list[float]]:

        """
        Generate ratings by iterating over each match, calculate predicted performance and update ratings after the match is finished.
        Default settin

        :param matches: list of matches. Each match must contain two teams.
        :param df: The dataframe from which the matches were generated. Only needed if you want to store the ratings in the class object in which case the column names must also be passed.
        :param column_names: The column names of the dataframe. Only needed if you want to store the ratings in the class object in which case the df must also be passed.
        :return: A dictionary containing historical match-rating values.
         These ratings can easily be added as new columns to the original dataframe for later model training or exploration
        """

        if self.column_names.participation_weight is not None and df is not None and self.column_names.participation_weight not in df.columns:
            raise ValueError(f"participation_weight {self.column_names.participation_weight} not in df columns")

        if matches is not None and len(matches) > 0 and not isinstance(matches[0], Match):
            raise ValueError("matches must be a list of Match objects")

        if matches is None and df is None:
            raise ValueError("If matches is not passed, df must be massed")

        if matches is None:
            matches = convert_df_to_matches(df=df, column_names=self.column_names)

        pre_match_player_rating_values = []
        pre_match_team_rating_values = []
        pre_match_opponent_projected_rating_values = []
        pre_match_opponent_rating_values = []
        team_opponent_leagues = []
        rating_update_match_ids = []
        rating_update_team_ids = []
        rating_update_team_ids_opponent = []
        player_rating_changes = []
        player_leagues = []
        player_predicted_performances = []
        projected_participation_weights = []
        pre_match_team_projected_rating_values = []
        position_rating_difference_values = {}
        performances = []
        player_ids = []
        team_leagues = []

        team_rating_changes = []

        for match_idx, match in enumerate(matches):
            self._validate_match(match)

            pre_match_rating = PreMatchRating(
                id=match.id,
                teams=self._get_pre_match_team_ratings(match=match),
                day_number=match.day_number
            )
            match_team_rating_changes = self._create_match_team_rating_changes(match=match,
                                                                               pre_match_rating=pre_match_rating)
            team_rating_changes += match_team_rating_changes

            if match_idx == len(matches) - 1 or matches[match_idx + 1].update_id != match.update_id:
                self._update_ratings(team_rating_changes=team_rating_changes)
                team_rating_changes = []

            match_position_ratings = []

            for team_idx, team_rating_change in enumerate(match_team_rating_changes):
                match_position_ratings.append({})
                opponent_team = match_team_rating_changes[-team_idx + 1]
                for player_idx, player_rating_change in enumerate(team_rating_change.players):

                    position = match.teams[team_idx].players[player_idx].position
                    if position:
                        match_position_ratings[team_idx][position] = player_rating_change.pre_match_rating_value

                    pre_match_team_projected_rating_values.append(
                        pre_match_rating.teams[team_idx].projected_rating_value)
                    pre_match_opponent_projected_rating_values.append(
                        pre_match_rating.teams[-team_idx + 1].projected_rating_value)

                    pre_match_player_rating_values.append(player_rating_change.pre_match_rating_value)
                    pre_match_team_rating_values.append(pre_match_rating.teams[team_idx].rating_value)
                    pre_match_opponent_rating_values.append(pre_match_rating.teams[-team_idx + 1].rating_value)
                    player_leagues.append(player_rating_change.league)
                    team_opponent_leagues.append(opponent_team.league)
                    team_leagues.append(team_rating_change.league)
                    rating_update_match_ids.append(match.update_id)
                    rating_update_team_ids.append(match.teams[team_idx].update_id)
                    rating_update_team_ids_opponent.append(match.teams[-team_idx + 1].update_id)
                    projected_participation_weights.append(
                        match.teams[team_idx].players[player_idx].performance.projected_participation_weight)

                    performances.append(player_rating_change.performance)
                    player_predicted_performances.append(player_rating_change.predicted_performance)
                    player_rating_changes.append(player_rating_change.rating_change_value)
                    player_ids.append(player_rating_change.id)

            if self.distinct_positions:
                for team_idx in range(len(match_team_rating_changes)):
                    player_per_team_count = len(match_team_rating_changes[team_idx].players)

                    for position in self.distinct_positions:
                        if position not in position_rating_difference_values:
                            position_rating_difference_values[position] = []
                        if position in match_position_ratings[team_idx] and position in match_position_ratings[
                            -team_idx + 1]:
                            position_rating_difference_values[position] += [match_position_ratings[team_idx][position] -
                                                                            match_position_ratings[-team_idx + 1][
                                                                                position]] * player_per_team_count
                        else:
                            position_rating_difference_values[position] += [0] * player_per_team_count

        potential_feature_values = self._get_shared_rating_values(
            position_rating_difference_values=position_rating_difference_values,
            pre_match_team_projected_rating_values=pre_match_team_projected_rating_values,
            pre_match_opponent_projected_rating_values=pre_match_opponent_projected_rating_values,
            pre_match_player_rating_values=pre_match_player_rating_values,
            player_leagues=player_leagues,
            team_opponent_leagues=team_opponent_leagues,
            projected_participation_weights=projected_participation_weights,
            match_ids=rating_update_match_ids,
            team_ids=rating_update_team_ids,
            team_id_opponents=rating_update_team_ids_opponent,
            player_ids=player_ids,
            team_leagues=team_leagues,
        )
        potential_feature_values[RatingHistoricalFeatures.PLAYER_RATING_DIFFERENCE] = np.array(
            pre_match_player_rating_values) - np.array(
            pre_match_opponent_rating_values)
        potential_feature_values[RatingHistoricalFeatures.RATING_DIFFERENCE] = np.array(
            pre_match_team_rating_values) - np.array(
            pre_match_opponent_rating_values)
        potential_feature_values[RatingEstimatorFeatures.PLAYER_RATING] = pre_match_player_rating_values
        potential_feature_values[RatingHistoricalFeatures.OPPONENT_RATING] = pre_match_opponent_rating_values
        potential_feature_values[RatingHistoricalFeatures.TEAM_RATING] = pre_match_team_rating_values
        potential_feature_values[RatingHistoricalFeatures.RATING_MEAN] = np.array(
            pre_match_team_rating_values) * 0.5 + 0.5 * np.array(pre_match_opponent_rating_values)

        potential_feature_values[RatingHistoricalFeatures.PLAYER_RATING_DIFFERENCE_FROM_TEAM] = np.array(
            pre_match_player_rating_values) - np.array(pre_match_team_rating_values)
        potential_feature_values[RatingHistoricalFeatures.PERFORMANCE] = performances

        potential_feature_values[RatingHistoricalFeatures.PLAYER_RATING_CHANGE] = player_rating_changes
        potential_feature_values[
            RatingHistoricalFeatures.PLAYER_PREDICTED_PERFORMANCE] = player_predicted_performances

        if df is not None and self.column_names:
            self.ratings_df = df[
                [self.column_names.team_id, self.column_names.player_id, self.column_names.match_id]].assign(
                **potential_feature_values)

        return {f: potential_feature_values[f] for f in self.estimator_features_return + self._historical_features_out}

    def generate_future(self, matches: Optional[list[Match]] = None, df: Optional[pd.DataFrame] = None) -> dict[Union[
                                                                                                                    RatingEstimatorFeatures, RatingHistoricalFeatures], list[float]]:

        if matches is not None and len(matches) > 0 and not isinstance(matches[0], Match):
            raise ValueError("matches must be a list of Match objects")

        if matches is None and df is None:
            raise ValueError("If matches is not passed, df must be massed")

        if matches is None:
            matches = convert_df_to_matches(df=df, column_names=self.column_names)

        pre_match_player_rating_values = []
        pre_match_opponent_projected_rating_values = []
        team_opponent_leagues = []
        match_ids = []
        player_leagues = []
        projected_participation_weights = []
        rating_update_team_ids = []
        rating_update_team_ids_opponent = []
        player_ids = []
        team_leagues = []
        position_rating_difference_values = {}

        pre_match_team_projected_rating_values = []

        for match_idx, match in enumerate(matches):
            match_position_ratings = []
            self._validate_match(match)
            pre_match_rating = PreMatchRating(
                id=match.id,
                teams=self._get_pre_match_team_ratings(match=match),
                day_number=match.day_number
            )

            for team_idx, pre_match_team in enumerate(pre_match_rating.teams):
                match_position_ratings.append({})
                opponent_team = pre_match_rating.teams[-team_idx + 1]
                for player_idx, pre_match_player in enumerate(pre_match_team.players):
                    position = match.teams[team_idx].players[player_idx].position
                    if position:
                        match_position_ratings[team_idx][position] = pre_match_player.rating_value
                    pre_match_team_projected_rating_values.append(pre_match_team.projected_rating_value)
                    pre_match_player_rating_values.append(pre_match_player.rating_value)
                    pre_match_opponent_projected_rating_values.append(opponent_team.projected_rating_value)
                    team_opponent_leagues.append(opponent_team.league)
                    team_leagues.append(pre_match_team.league)
                    player_leagues.append(pre_match_player.league)
                    match_ids.append(match.id)
                    player_ids.append(pre_match_player.id)
                    rating_update_team_ids.append(match.teams[team_idx].update_id)
                    rating_update_team_ids_opponent.append(match.teams[-team_idx + 1].update_id)
                    projected_participation_weights.append(
                        match.teams[team_idx].players[player_idx].performance.projected_participation_weight)

            if self.distinct_positions:
                for team_idx in range(len(pre_match_rating.teams)):
                    player_per_team_count = len(pre_match_rating.teams[team_idx].players)

                    for position in self.distinct_positions:
                        if position not in position_rating_difference_values:
                            position_rating_difference_values[position] = []
                        if position in match_position_ratings[team_idx] and position in match_position_ratings[
                            -team_idx + 1]:
                            position_rating_difference_values[position] += [match_position_ratings[team_idx][position] -
                                                                            match_position_ratings[-team_idx + 1][
                                                                                position]] * player_per_team_count
                        else:
                            position_rating_difference_values[position] += [0] * player_per_team_count

        potential_feature_values = self._get_shared_rating_values(
            position_rating_difference_values=position_rating_difference_values,
            pre_match_team_projected_rating_values=pre_match_team_projected_rating_values,
            pre_match_opponent_projected_rating_values=pre_match_opponent_projected_rating_values,
            pre_match_player_rating_values=pre_match_player_rating_values,
            player_leagues=player_leagues,
            team_opponent_leagues=team_opponent_leagues,
            match_ids=match_ids,
            projected_participation_weights=projected_participation_weights,
            team_ids=rating_update_team_ids,
            team_id_opponents=rating_update_team_ids_opponent,
            player_ids=player_ids,
            team_leagues=team_leagues,
        )

        return {f: potential_feature_values[f] for f in self.estimator_features_return + self._historical_features_out}

    def _get_shared_rating_values(self,
                                  position_rating_difference_values: dict[str, list[float]],
                                  pre_match_team_projected_rating_values: list[float],
                                  pre_match_opponent_projected_rating_values: list[float],
                                  pre_match_player_rating_values: list[float],
                                  player_leagues: list[str],
                                  team_opponent_leagues: list[str],
                                  match_ids: list[str],
                                  team_ids: list[str],
                                  team_id_opponents: list[str],
                                  player_ids: list[str],
                                  projected_participation_weights: list[float],
                                  team_leagues: list[str],
                                  ) -> dict[Union[RatingEstimatorFeatures, RatingHistoricalFeatures], Any]:

        if self.column_names.projected_participation_weight:
            df = pd.DataFrame({
                "match_id": match_ids,
                "team_id": team_ids,
                "team_id_opponent": team_id_opponents,
                RatingEstimatorFeatures.PLAYER_RATING: pre_match_player_rating_values,
                "projected_participation_weight": projected_participation_weights,
                "player_id": player_ids,
            })

            game_player = df.groupby(["match_id", "player_id", "team_id", "team_id_opponent"])[
                [RatingEstimatorFeatures.PLAYER_RATING, "projected_participation_weight"]].mean().reset_index()

            game_player["game_team_sum_projected_participation_weight"] = game_player.groupby(["match_id", "team_id"])[
                "projected_participation_weight"].transform('sum')

            game_player['weighted_pre_match_player_rating_value'] = game_player[RatingEstimatorFeatures.PLAYER_RATING] * \
                                                                    game_player["projected_participation_weight"]

            game_player[RatingEstimatorFeatures.TEAM_RATING_PROJECTED] = game_player.groupby(["match_id", "team_id"])[
                                                                       "weighted_pre_match_player_rating_value"].transform(
                'sum') / game_player['game_team_sum_projected_participation_weight']

            game_team = game_player.groupby(["match_id", "team_id", "team_id_opponent"])[
                RatingEstimatorFeatures.TEAM_RATING_PROJECTED].mean().reset_index()

            game_team = game_team.merge(
                game_team[["match_id", "team_id_opponent", RatingEstimatorFeatures.TEAM_RATING_PROJECTED]].rename(
                    columns={RatingEstimatorFeatures.TEAM_RATING_PROJECTED: RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED}),
                left_on=["match_id", "team_id"], right_on=["match_id", "team_id_opponent"])

            game_player = game_player.merge(
                game_team[['match_id', 'team_id', RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED]],
                on=["match_id", "team_id"])

            game_player[RatingEstimatorFeatures.RATING_MEAN_PROJECTED] = (game_player[
                                                                        RatingEstimatorFeatures.TEAM_RATING_PROJECTED] +
                                                                          game_player[
                                                                        RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED]) / 2

            df = df[["match_id", "player_id"]].merge(
                game_player[["match_id", "player_id",
                             RatingEstimatorFeatures.TEAM_RATING_PROJECTED,
                             RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED,
                             RatingEstimatorFeatures.RATING_MEAN_PROJECTED,
                             RatingEstimatorFeatures.PLAYER_RATING
                             ]], on=["match_id", "player_id"], how='left')

            rating_differences_projected = (df[RatingEstimatorFeatures.TEAM_RATING_PROJECTED] - df[
                RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED]).tolist()
            player_rating_difference_from_team_projected = (
                    df[RatingEstimatorFeatures.PLAYER_RATING] - df[RatingEstimatorFeatures.TEAM_RATING_PROJECTED]).tolist()
            player_rating_differences_projected = (
                    df[RatingEstimatorFeatures.PLAYER_RATING] - df[RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED]).tolist()
            rating_means_projected = df[RatingEstimatorFeatures.RATING_MEAN_PROJECTED].tolist()
            pre_match_opponent_projected_rating_values = df[RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED].tolist()
            pre_match_team_projected_rating_values = df[RatingEstimatorFeatures.TEAM_RATING_PROJECTED].tolist()
            pre_match_player_rating_values = df[RatingEstimatorFeatures.PLAYER_RATING].tolist()

        else:
            rating_differences_projected = (np.array(pre_match_team_projected_rating_values) - np.array(
                pre_match_opponent_projected_rating_values)).tolist()
            player_rating_differences_projected = (np.array(pre_match_player_rating_values) - np.array(
                pre_match_opponent_projected_rating_values)).tolist()
            player_rating_difference_from_team_projected = (np.array(pre_match_player_rating_values) - np.array(
                pre_match_team_projected_rating_values)).tolist()
            rating_means_projected = (np.array(pre_match_team_projected_rating_values) * 0.5 + 0.5 * np.array(
                pre_match_opponent_projected_rating_values)).tolist()

        return_values = {
            RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED: rating_differences_projected,
            RatingEstimatorFeatures.PLAYER_RATING_DIFFERENCE_FROM_TEAM_PROJECTED: player_rating_difference_from_team_projected,
            RatingEstimatorFeatures.PLAYER_RATING_DIFFERENCE_PROJECTED: player_rating_differences_projected,
            RatingEstimatorFeatures.TEAM_RATING_PROJECTED: pre_match_team_projected_rating_values,
            RatingEstimatorFeatures.OPPONENT_RATING_PROJECTED: pre_match_opponent_projected_rating_values,
            RatingEstimatorFeatures.PLAYER_RATING: pre_match_player_rating_values,
            RatingEstimatorFeatures.PLAYER_LEAGUE: player_leagues,
            RatingEstimatorFeatures.OPPONENT_LEAGUE: team_opponent_leagues,
            RatingEstimatorFeatures.TEAM_LEAGUE: team_leagues,
            RatingEstimatorFeatures.RATING_MEAN_PROJECTED: rating_means_projected,
            RatingEstimatorFeatures.MATCH_ID: match_ids,
        }

        if self.distinct_positions:
            for position, rating_values in position_rating_difference_values.items():
                return_values[RatingEstimatorFeatures.RATING_DIFFERENCE_POSITION + "_" + position] = rating_values

        return return_values

    def _create_match_team_rating_changes(self, match: Match, pre_match_rating: PreMatchRating) -> list[
        TeamRatingChange]:

        team_rating_changes = []

        for team_idx, pre_match_team_rating in enumerate(pre_match_rating.teams):
            team_rating_change = self.team_rating_generator.generate_rating_change(day_number=match.day_number,
                                                                                   pre_match_team_rating=pre_match_team_rating,
                                                                                   pre_match_opponent_team_rating=
                                                                                   pre_match_rating.teams[
                                                                                       -team_idx + 1])
            team_rating_changes.append(team_rating_change)

        return team_rating_changes

    def _update_ratings(self, team_rating_changes: list[TeamRatingChange]):

        for idx, team_rating_change in enumerate(team_rating_changes):
            self.team_rating_generator.update_rating_by_team_rating_change(team_rating_change=team_rating_change,
                                                                           opponent_team_rating_change=
                                                                           team_rating_changes[-idx + 1])

    def _get_pre_match_team_ratings(self, match: Match) -> list[PreMatchTeamRating]:
        pre_match_team_ratings = []
        for match_team in match.teams:
            pre_match_team_ratings.append(self.team_rating_generator.generate_pre_match_team_rating(
                match_team=match_team, day_number=match.day_number))

        return pre_match_team_ratings

    def _validate_match(self, match: Match):
        if len(match.teams) < 2:
            raise ValueError(f"{match.id} only contains {len(match.teams)} teams")

    @property
    def player_ratings(self) -> dict[str, PlayerRating]:
        return dict(sorted(self.team_rating_generator.player_ratings.items(),
                           key=lambda item: item[1].rating_value, reverse=True))

    @property
    def team_ratings(self) -> list[TeamRating]:
        team_id_ratings: list[TeamRating] = []
        teams = self.team_rating_generator.teams
        player_ratings = self.player_ratings
        for id, team in teams.items():
            team_player_ratings = [player_ratings[p] for p in team.player_ids]
            team_rating_value = sum([p.rating_value for p in team_player_ratings]) / len(team_player_ratings)
            team_id_ratings.append(TeamRating(id=team.id, name=team.name, players=team_player_ratings,
                                              last_match_day_number=team.last_match_day_number,
                                              rating_value=team_rating_value))

        return list(sorted(team_id_ratings,
                           key=lambda team: team.rating_value, reverse=True))

    @property
    def estimator_features_out(self) -> list[RatingEstimatorFeatures]:
        return self._estimator_features_out

    @property
    def features_out(self) -> list[Union[RatingEstimatorFeatures, RatingHistoricalFeatures]]:
        return self._estimator_features_out + self._historical_features_out

    @property
    def estimator_features_return(self) -> list[RatingEstimatorFeatures]:
        if self._estimator_features_pass_through:
            return list(set(self._estimator_features_pass_through + self.estimator_features_out))
        return self.estimator_features_out


from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from player_performance_ratings.scorer.score import Filter

class PredictorTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


class BaseMLWrapper(ABC):

    def __init__(self,
                 estimator,
                 filters: Optional[list[Filter]],
                 target: str,
                 categorical_transformers: Optional[list[PredictorTransformer]] = None,
                 pred_column: Optional[str] = None,
                 ):
        self.estimator = estimator
        self.filters = filters or []
        self._target = target
        self._pred_column = pred_column or f"{self._target}_prediction"
        self.categorical_transformers = categorical_transformers
        self._estimator_categorical_features = []
        self._deepest_estimator = self.estimator
        iterations = 0
        while hasattr(self._deepest_estimator , "estimator"):
            self._deepest_estimator  = self._deepest_estimator.estimator
            iterations += 1
            if iterations > 10:
                raise ValueError("estimator is too deep")

    @abstractmethod
    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        pass

    @abstractmethod
    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def pred_column(self) -> str:
        return self._pred_column

    @property
    def target(self) -> str:
        return self._target

    @property
    def classes_(self) -> Optional[list[str]]:
        if 'classes_' not in dir(self.estimator):
            return None
        return self.estimator.classes_

    def set_target(self, new_target_name: str):
        self._target = new_target_name

    def fit_transform_categorical_transformers(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.categorical_transformers:
            for pre_transformer in self.categorical_transformers:
                df = pre_transformer.fit_transform(df)
                for feature in pre_transformer.features:
                    if feature in self._estimator_features:
                        self._estimator_features.remove(feature)
                self._estimator_features = list(set(pre_transformer.features_out + self._estimator_features))
                self._estimator_categorical_features = list(
                    set(pre_transformer.features_out + self._estimator_categorical_features))
        return df

    def transform_categorical_transformers(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.categorical_transformers:
            for pre_transformer in self.categorical_transformers:
                df = pre_transformer.transform(df)
        return df

    @property
    def estimator_features(self) -> list[str]:
        return self._estimator_features

    @property
    def estimator_categorical_features(self) -> list[str]:
        return self._estimator_categorical_features



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

import numpy as np
import pandas as pd

from player_performance_ratings import ColumnNames
from player_performance_ratings.predictor import Predictor, BaseMLWrapper
from player_performance_ratings.transformation.base_transformer import DifferentGranularityTransformer, \
    BasePostTransformer
from player_performance_ratings.utils import validate_sorting


def create_output_column_by_game_group(data: pd.DataFrame, feature_name: str,
                                       weight_column: str, game_id: str, granularity: list[str]) -> pd.DataFrame:
    if weight_column:
        data = data.assign(**{f'weighted_{feature_name}': data[feature_name] * data[weight_column]})
        data = data.assign(
            **{"sum_weighted": data.groupby(granularity + [game_id])[weight_column].transform("sum")})
        data = data.assign(**{feature_name: data[f'weighted_{feature_name}'] / data['sum_weighted']})
        data = data.groupby(granularity + [game_id])[feature_name].sum().reset_index()

    else:
        data = data.groupby(granularity + [game_id])[feature_name].mean().reset_index()
    return data


class PredictorTransformer(BasePostTransformer):

    def __init__(self, predictor: BaseMLWrapper, features: list[str] = None):
        self.predictor = predictor
        super().__init__(features=features)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.predictor.train(df=df, estimator_features=self.features)
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)
        return df

    @property
    def features_out(self) -> list[str]:
        return [f'{self.predictor.pred_column}']


class RatioTeamPredictorTransformer(BasePostTransformer):
    def __init__(self,
                 features: list[str],
                 predictor: BaseMLWrapper,
                 game_id: str,
                 team_id: str,
                 team_total_prediction_column: Optional[str] = None,
                 prefix: str = "_ratio_team"
                 ):
        super().__init__(features=features)
        self.predictor = predictor
        self.game_id = game_id
        self.team_id = team_id
        self.team_total_prediction_column = team_total_prediction_column
        self.prefix = prefix
        self.predictor._pred_column = f"__prediction__{self.predictor.target}"
        self._features_out = [self.predictor.target + prefix]
        if self.team_total_prediction_column:
            self._features_out.append(self.predictor.target + prefix + "_team_total_multiplied")

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.predictor.train(df=df, estimator_features=self.features)
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)
        df[self.predictor.pred_column + "_sum"] = df.groupby([self.game_id, self.team_id])[
            self.predictor.pred_column].transform('sum')
        df[self._features_out[0]] = df[self.predictor.pred_column] / df[self.predictor.pred_column + "_sum"]
        if self.team_total_prediction_column:
            df = df.assign(**{self.predictor.target + self.prefix + "_team_total_multiplied": df[self._features_out[
                0]] * df[
                                                                                                  self.team_total_prediction_column]})
        return df.drop(columns=[self.predictor.pred_column + "_sum", self.predictor.pred_column])

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class SklearnPredictorTransformer(BasePostTransformer):

    def __init__(self, estimator, features: list[str], target: str, train_date: str, date_column_name: str,
                 feature_out: Optional[str] = None):
        super().__init__(features=features)
        self.estimator = estimator
        self.target = target
        self.train_date = train_date
        self.date_column_name = date_column_name
        self.feature_out = feature_out or f'{self.target}_transform_prediction'

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        train_df = df[df[self.date_column_name] < self.train_date]
        self.estimator.fit(train_df[self.features], train_df[self.target])
        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        prediction = self.estimator.predict(df[self.features])
        df = df.assign(**{self.feature_out: prediction})
        return df

    @property
    def features_out(self) -> list[str]:
        return [f'{self.feature_out}']


class GameTeamMembersColumnsTransformer(BasePostTransformer):

    def __init__(self, column_names: ColumnNames, features: list[str], players_per_match_per_team_count: int,
                 sort_by: Optional[list[str]] = None):
        super().__init__(features=features)
        self.column_names = column_names
        self.players_per_match_per_team_count = players_per_match_per_team_count
        self.sort_by = sort_by
        self._features_out = []
        for number in range(1, self.players_per_match_per_team_count + 1):
            for feature in self.features:
                feature_name = f'team_player{number}_{feature}'
                self._features_out.append(feature_name)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.copy()

        def add_teammate_minutes(row, grouped_df):
            game_team_group = grouped_df.get_group(
                (row[self.column_names.rating_update_match_id], row[self.column_names.team_id]))
            number = 0
            if self.sort_by:
                game_team_group = game_team_group.sort_values(by=self.sort_by, ascending=False)

            for index, teammate in game_team_group.iterrows():
                if teammate[self.column_names.player_id] != row[self.column_names.player_id]:
                    number += 1
                    for feature in self.features:
                        feature_name = f'team_player{number}_{feature}'
                        row[feature_name] = teammate[feature]

            return row

        grouped = df.groupby([self.column_names.rating_update_match_id, self.column_names.team_id])

        df = df.apply(lambda row: add_teammate_minutes(row, grouped), axis=1)
        return df.sort_values(
            by=[self.column_names.start_date, self.column_names.rating_update_match_id, self.column_names.team_id,
                self.column_names.player_id])

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    @property
    def features_out(self) -> list[str]:
        return self.features


class NormalizerTargetColumnTransformer(BasePostTransformer):

    def __init__(self, features: list[str], granularity, target_sum_column_name: str, prefix: str = "__normalized_"):
        super().__init__(features=features)
        self.granularity = granularity
        self.prefix = prefix
        self.target_sum_column_name = target_sum_column_name
        self._features_to_normalization_target = {}
        self._features_out = []
        for feature in self.features:
            self._features_out.append(f'{self.prefix}{feature}')

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            df[f"{feature}_sum"] = df.groupby(self.granularity)[feature].transform('sum')
            df = df.assign(
                **{self.prefix + feature: df[feature] / df[f"{feature}_sum"] * df[self.target_sum_column_name]})
            df = df.drop(columns=[f"{feature}_sum"])
        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class NormalizerTransformer(BasePostTransformer):

    def __init__(self, features: list[str], granularity, target_mean: Optional[float] = None,
                 create_target_as_mean: bool = False):
        super().__init__(features=features)
        self.granularity = granularity
        self.target_mean = target_mean
        self.create_target_as_mean = create_target_as_mean
        self._features_to_normalization_target = {}

        if self.target_mean is None and not self.create_target_as_mean:
            raise ValueError("Either target_sum or create_target_as_mean must be set")

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._features_to_normalization_target = {f: self.target_mean for f in self.features} if self.target_mean else {
            f: df[f].mean() for f in self.features}
        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{f'__mean_{feature}': df.groupby(self.granularity)[feature].transform('mean') for feature in
                          self.features})
        for feature, target_sum in self._features_to_normalization_target.items():
            df = df.assign(**{feature: df[feature] / df[f'__mean_{feature}'] * target_sum})
        return df.drop(columns=[f'__mean_{feature}' for feature in self.features])

    @property
    def features_out(self) -> list[str]:
        return self.features


class LagTransformer(BasePostTransformer):

    def __init__(self,
                 features: list[str],
                 lag_length: int,
                 column_names: ColumnNames,
                 granularity: Optional[list[str]] = None,
                 days_between_lags: Optional[list[int]] = None,
                 prefix: str = 'lag_',
                 future_lag: bool = False,
                 ):

        """
        :param
            feature_names: Which features to lag

        :param lag_length:
            Number of lags

        :param granularity:
            Columns to group by before lagging. E.g. player_id or [player_id, position].
            In the latter case it will get the lag for each player_id and position combination.
            Defaults to player_id

        :param column_names:


        :param group_to_game_level
            If there are more multiple rows per granularity per game_id and you want to add the lag from the prior games, set to True.
            This will calculate the mean of the features per game and the lag will be the prior means


        :param prefix:
            Prefix for the new lag columns
        """

        super().__init__(features=features)
        self.column_names = column_names
        self.granularity = granularity or [self.column_names.player_id]
        self.lag_length = lag_length
        self.days_between_lags = days_between_lags or []
        self.prefix = prefix
        self.future_lag = future_lag
        self._features_out = []
        self._df = None
        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                self._features_out.append(f'{self.prefix}{lag}_{feature_name}')

        for days_lag in self.days_between_lags:
            self._features_out.append(f'{self.prefix}{days_lag}_days_ago')

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})

        for feature_out in self._features_out:
            if feature_out in df.columns:
                raise ValueError(
                    f'Column {feature_out} already exists. Choose different prefix or ensure no duplication was performed')
        if self._df is None:
            self._df = df
        else:
            self._df = pd.concat([self._df, df], axis=0)

        self._df = self._df.assign(
            __id=self._df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                           self.column_names.player_id]].agg('__'.join, axis=1))
        self._df = self._df.drop_duplicates(subset=['__id'], keep='last')

        transformed_df = self.transform(pd.DataFrame(df))
        return transformed_df

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})
        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        if len(df.drop_duplicates(
                subset=[self.column_names.player_id, self.column_names.parent_team_id,
                        self.column_names.rating_update_match_id])) != len(df):
            raise ValueError(
                f"Duplicated rows in df. Df must be a unique combination of {self.column_names.player_id} and {self.column_names.rating_update_match_id}")

        ori_cols = df.columns.tolist()
        ori_index_values = df.index.tolist()

        all_df = pd.concat([self._df, df], axis=0)
        all_df = all_df.assign(
            __id=all_df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                         self.column_names.player_id]].agg('__'.join, axis=1))
        all_df = all_df.drop_duplicates(subset=['__id'], keep='last')
        if self.column_names.participation_weight:
            for feature in self.features:
                all_df = all_df.assign(**{feature: all_df[feature] * all_df[self.column_names.participation_weight]})

        grouped = \
            all_df.groupby(self.granularity + [self.column_names.rating_update_match_id, self.column_names.start_date])[
                self.features].mean().reset_index()

        grouped = grouped.sort_values([self.column_names.start_date, self.column_names.rating_update_match_id])

        for days_lag in self.days_between_lags:
            if self.future_lag:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[self.column_names.start_date].shift(
                    -days_lag)
                grouped[f'{self.prefix}{days_lag}_days_ago'] = (
                        pd.to_datetime(grouped["shifted_days"]) - pd.to_datetime(
                    grouped[self.column_names.start_date])).dt.days
            else:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[self.column_names.start_date].shift(
                    days_lag)
                grouped[f'{self.prefix}{days_lag}_days_ago'] = (pd.to_datetime(
                    grouped[self.column_names.start_date]) - pd.to_datetime(grouped["shifted_days"])).dt.days

            grouped = grouped.drop(columns=["shifted_days"])

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f'{self.prefix}{lag}_{feature_name}'
                if output_column_name in all_df.columns:
                    raise ValueError(
                        f'Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed')

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f'{self.prefix}{lag}_{feature_name}'
                if self.future_lag:
                    grouped = grouped.assign(
                        **{output_column_name: grouped.groupby(self.granularity)[feature_name].shift(-lag)})
                else:
                    grouped = grouped.assign(
                        **{output_column_name: grouped.groupby(self.granularity)[feature_name].shift(lag)})

        all_df = all_df.merge(
            grouped[self.granularity + [self.column_names.rating_update_match_id, *self.features_out]],
            on=self.granularity + [self.column_names.rating_update_match_id], how='left')

        df = df.assign(
            __id=df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))

        transformed_df = all_df[all_df['__id'].isin(df['__id'].unique().tolist())][ori_cols + self._features_out]
        transformed_df.index = ori_index_values
        return transformed_df[list(set(ori_cols + self._features_out))]

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class LagLowerGranularityTransformer(DifferentGranularityTransformer):
    """
    TODO: Make work
    """

    def __init__(self,
                 features: list[str],
                 lag_length: int,
                 column_names: ColumnNames,
                 granularity: Union[list[str], str] = None,
                 weight_column: Optional[str] = None,
                 prefix: str = 'lag_lower_'
                 ):
        """
        :param
            feature_names: Which features to lag

        :param lag_length:
            Number of lags

        :param granularity:
            Columns to group by before lagging. E.g. player_id or [player_id, position].
            In the latter case it will get the lag for each player_id and position combination.
            Defaults to player_id

        :param column_names:
        """
        super().__init__(features)
        self.column_names = column_names
        self.granularity = granularity or [self.column_names.player_id]
        self.lag_length = lag_length
        self.prefix = prefix
        self.weight_column = weight_column
        self._features_out = []
        self._diff_granularity_df = None
        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                self._features_out.append(f'{self.prefix}{lag}_{feature_name}')

    def fit_transform(self, diff_granularity_df: pd.DataFrame, game_player_df: pd.DataFrame) -> pd.DataFrame:

        validate_sorting(df=game_player_df, column_names=self.column_names)

        if self._diff_granularity_df is None:
            self._diff_granularity_df = diff_granularity_df
        else:
            self._diff_granularity_df = pd.concat([self._diff_granularity_df, diff_granularity_df], axis=0)

        if len(self._diff_granularity_df.drop_duplicates()) != len(self._diff_granularity_df):
            raise ValueError(
                "Duplicated rows in diff_granularity_df. Please ensure there are no duplicates in the data")

        empty_diff_granularity_df = pd.DataFrame(columns=diff_granularity_df.columns)

        transformed_game_player_df = self.transform(diff_granularity_df=empty_diff_granularity_df,
                                                    game_player_df=game_player_df)

        return transformed_game_player_df

    def transform(self, diff_granularity_df: pd.DataFrame, game_player_df: pd.DataFrame) -> pd.DataFrame:
        validate_sorting(df=game_player_df, column_names=self.column_names)
        if self._diff_granularity_df is None:
            raise ValueError("fit_transform needs to be called before transform")

        diff_granularity_df = pd.concat([self._diff_granularity_df, diff_granularity_df], axis=0)

        diff_granularity_df = diff_granularity_df.sort_values(
            by=[self.column_names.start_date, self.column_names.rating_update_match_id, self.column_names.team_id,
                self.column_names.player_id])

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f'{self.prefix}{lag}_{feature_name}'
                if output_column_name in diff_granularity_df.columns:
                    raise ValueError(
                        f'Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed')

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                grouped_data = create_output_column_by_game_group(data=diff_granularity_df, feature_name=feature_name,
                                                                  weight_column=self.weight_column,
                                                                  game_id=self.column_names.rating_update_match_id,
                                                                  granularity=self.granularity)

                output_column_name = f'{self.prefix}{lag}_{feature_name}'

                grouped_data = grouped_data.assign(
                    **{output_column_name: grouped_data.groupby(self.granularity)[feature_name].shift(lag)})
                game_player_df = game_player_df[[c for c in game_player_df.columns if c != output_column_name]].merge(
                    grouped_data[[output_column_name, self.column_names.rating_update_match_id] + self.granularity],
                    on=self.granularity + [self.column_names.rating_update_match_id], how='left')

        return game_player_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanTransformer(BasePostTransformer):

    def __init__(self,
                 features: list[str],
                 window: int,
                 column_names: ColumnNames,
                 granularity: Union[list[str], str] = None,
                 min_periods: int = 1,
                 prefix: str = 'rolling_mean_'):
        """

        :param features:
            Features to create rolling mean for

        :param granularity:
            Columns to group by before rolling mean. E.g. player_id or [player_id, position].
             In the latter case it will get the rolling mean for each player_id and position combination.

        :param window:
            Window size for rolling mean, if 10 will calculate rolling mean over the prior 10 observations

        :param game_id:
            Column name of game_id.
            If there are more multiple rows per granularity per game_id and you want to add the rolling mean per game_id, set game_id
            This will calculate the mean of the features per game and the rolling mean will be calculated on that.

        :param weight_column:
            Only used if game_id is set.
            Will calculate weighted mean of the features per game and the rolling mean will be the prior weighted means.
            This is useful when working with partial game-data of different lengths.
            In that case it can beneficial to set the game-length as weight_colum.

        :param df: Optional parameter to pass in a dataframe to calculate the rolling mean on.
            If not passed in, it will use the dataframe passed in the transform method.
            This is useful if you want to calculate the rolling mean on a different dataframe than the one you want to transform.
            Will merge the two dataframes on game_id and granularity.

        :param prefix:
            Prefix for the new rolling mean columns
        """

        super().__init__(features=features)
        self.granularity = granularity or [column_names.player_id]
        if isinstance(self.granularity, str):
            self.granularity = [self.granularity]
        self.window = window
        self.min_periods = min_periods
        self.column_names = column_names
        self._df = None
        self.prefix = prefix
        self._features_out = [f'{self.prefix}{self.window}_{c}' for c in self.features]

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})
        validate_sorting(df=df, column_names=self.column_names)
        if self._df is None:
            self._df = df
        else:
            self._df = pd.concat([self._df, df], axis=0)

        self._df = self._df.assign(
            __id=self._df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                           self.column_names.player_id]].agg('__'.join, axis=1))
        self._df = self._df.drop_duplicates(subset=['__id'], keep='last')

        transformed_df = self.transform(pd.DataFrame(df))
        return transformed_df

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})
        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        ori_cols = df.columns.tolist()
        ori_index_values = df.index.tolist()

        all_df = pd.concat([self._df, df], axis=0).reset_index()
        all_df = all_df.assign(
            __id=all_df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                         self.column_names.player_id]].agg('__'.join, axis=1))
        all_df = all_df.drop_duplicates(subset=['__id'], keep='last')

        for feature_name in self.features:

            if self.column_names.participation_weight:
                all_df = all_df.assign(
                    **{feature_name: all_df[feature_name] * all_df[self.column_names.participation_weight]})

            output_column_name = f'{self.prefix}{self.window}_{feature_name}'
            if output_column_name in all_df.columns:
                raise ValueError(
                    f'Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed')

            agg_dict = {feature_name: 'mean', self.column_names.start_date: 'first'}
            grp = all_df.groupby(self.granularity + [self.column_names.rating_update_match_id]).agg(
                agg_dict).reset_index()
            grp.sort_values(by=[self.column_names.start_date, self.column_names.rating_update_match_id], inplace=True)

            grp = grp.assign(**{output_column_name: grp.groupby(self.granularity)[feature_name].apply(
                lambda x: x.shift().rolling(self.window, min_periods=self.min_periods).mean())})

            all_df = all_df.merge(
                grp[self.granularity + [self.column_names.rating_update_match_id, output_column_name]],
                on=self.granularity + [self.column_names.rating_update_match_id], how='left')
            all_df = all_df.sort_values(by=[self.column_names.start_date, self.column_names.match_id,
                                            self.column_names.team_id, self.column_names.player_id])
        df = df.assign(
            __id=df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))
        transformed_df = all_df[all_df['__id'].isin(df['__id'].unique().tolist())][ori_cols + self._features_out]
        transformed_df.index = ori_index_values
        return transformed_df[list(set(ori_cols + self._features_out))]

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanDaysTransformer(BasePostTransformer):

    def __init__(self,
                 features: list[str],
                 days: Union[int, list[int]],
                 column_names: ColumnNames,
                 granularity: Union[list[str], str] = None,
                 add_count: bool = False,
                 prefix: str = 'rolling_mean_days_'):
        super().__init__(features=features)
        self.features = features
        self.days = days
        if isinstance(self.days, int):
            self.days = [self.days]
        self.column_names = column_names
        self.granularity = granularity or [self.column_names.player_id]
        self.add_count = add_count
        self.prefix = prefix
        self._features_out = []
        self._df = None

        for day in self.days:
            for feature_name in self.features:
                self._features_out.append(f'{self.prefix}{day}_{feature_name}')
            if self.add_count:
                self._features_out.append(f'{self.prefix}{day}_count')

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})

        validate_sorting(df=df, column_names=self.column_names)
        if self._df is None:
            self._df = df
        else:
            self._df = pd.concat([self._df, df], axis=0)

        self._df = self._df.assign(
            __id=self._df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                           self.column_names.player_id]].agg('__'.join, axis=1))
        self._df = self._df.drop_duplicates(subset=['__id'], keep='last')

        transformed_df = self.transform(df)
        return transformed_df

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        df = df.assign(**{self.column_names.player_id: lambda x: x[self.column_names.player_id].astype('str')})
        df = df.assign(**{
            self.column_names.rating_update_match_id: lambda x: x[self.column_names.rating_update_match_id].astype(
                'str')})
        df = df.assign(**{
            self.column_names.parent_team_id: lambda x: x[self.column_names.parent_team_id].astype(
                'str')})
        df = df.assign(
            __id=df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))

        all_df = pd.concat([self._df, df], axis=0).reset_index()
        all_df = all_df.drop_duplicates(subset=['__id'], keep='last')

        ori_cols = df.columns.tolist()
        ori_index_values = df.index.tolist()

        for feature_name in self.features:

            if self.column_names.participation_weight:
                all_df = all_df.assign(
                    **{feature_name: all_df[feature_name] * all_df[self.column_names.participation_weight]})


        all_df[self.column_names.start_date] = pd.to_datetime(all_df[self.column_names.start_date]).dt.date

        for day in self.days:

            df1 = (all_df
                   .groupby([self.column_names.start_date, *self.granularity])[self.features]
                   .agg(['sum', 'size'])
                   .unstack()
                   .asfreq('d', fill_value=np.nan)
                   .rolling(window=day, min_periods=1)
                   .sum()
                   .shift()
                   .stack()
                   )
            feats = []
            for feature_name in self.features:
                feats.append(f'{self.prefix}{day}_{feature_name}_sum')
                feats.append(f'{self.prefix}{day}_{feature_name}_count')

            df1.columns = feats
            for feature_name in self.features:
                df1[f'{self.prefix}{day}_{feature_name}'] = df1[f'{self.prefix}{day}_{feature_name}_sum'] / df1[
                    f'{self.prefix}{day}_{feature_name}_count']

            if self.add_count:
                df1[f'{self.prefix}{day}_count'] = df1[f'{self.prefix}{day}_{self.features[0]}_count']
                df1 = df1.drop(columns=[f'{self.prefix}{day}_{feature_name}_count' for feature_name in self.features])

            all_df[self.column_names.start_date] = pd.to_datetime(all_df[self.column_names.start_date])
            all_df = all_df.join(df1[[c for c in df1.columns if c in self.features_out]],
                                 on=[self.column_names.start_date, *self.granularity])
            if self.add_count:
                all_df[f'{self.prefix}{day}_count'] = all_df[f'{self.prefix}{day}_count'].fillna(0)

        all_df = all_df.sort_values(by=[self.column_names.start_date, self.column_names.match_id,
                                        self.column_names.team_id, self.column_names.player_id])

        df = df.assign(
            __id=df[[self.column_names.rating_update_match_id, self.column_names.parent_team_id,
                     self.column_names.player_id]].agg('__'.join, axis=1))
        transformed_df = all_df[all_df['__id'].isin(df['__id'].unique().tolist())][ori_cols + self._features_out]
        transformed_df.index = ori_index_values
        return transformed_df[list(set(ori_cols + self._features_out))].drop(columns=['__id'])

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class Operation(Enum):
    SUBTRACT = "subtract"


@dataclass
class ModifyOperation:
    feature1: str
    operation: Operation
    feature2: str
    new_column_name: Optional[str] = None

    def __post_init__(self):
        if self.operation == Operation.SUBTRACT and not self.new_column_name:
            self.new_column_name = f"{self.feature1}_minus_{self.feature2}"


class ModifierTransformer(BasePostTransformer):

    def __init__(self,
                 modify_operations: list[ModifyOperation],
                 features: list[str] = None,
                 are_estimator_features: bool = True,
                 ):
        super().__init__(features=features, are_estimator_features=are_estimator_features)
        self.modify_operations = modify_operations
        self._features_out = [operation.new_column_name for operation in self.modify_operations]

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for operation in self.modify_operations:
            if operation.operation == Operation.SUBTRACT:
                df[operation.new_column_name] = df[operation.feature1] - df[operation.feature2]

        return df



import pandas as pd
from lightgbm import LGBMRegressor
from sklearn.metrics import mean_absolute_error

from player_performance_ratings.scorer.score import SklearnScorer
from player_performance_ratings.tuner.rating_generator_tuner import UpdateRatingGeneratorTuner

from player_performance_ratings import ColumnNames, PredictColumnNames
from player_performance_ratings.predictor.estimators import Predictor
from player_performance_ratings.ratings import RatingEstimatorFeatures, OpponentAdjustedRatingGenerator, ColumnWeight
from player_performance_ratings.tuner import PipelineTuner
from player_performance_ratings.tuner.match_predictor_factory import PipelineFactory
from player_performance_ratings.tuner.utils import get_default_team_rating_search_range

df = pd.read_pickle(r"data/game_player_subsample.pickle")

df = (
    df.assign(team_count=df.groupby("game_id")["team_id"].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)

df = df.sort_values(["start_date", "game_id", "team_id", "player_id"])
df[PredictColumnNames.TARGET] = df['points']

column_names = ColumnNames(
    match_id="game_id",
    team_id="team_id",
    player_id="player_id",
    start_date="start_date",
    performance="points"
)

match_predictor_factory = PipelineFactory(
    rating_generators=[OpponentAdjustedRatingGenerator(column_names=column_names,
                                                       features_out=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED,
                                                                     RatingEstimatorFeatures.PLAYER_RATING_DIFFERENCE_PROJECTED])],
    predictor=Predictor(estimator=LGBMRegressor(), features=[RatingEstimatorFeatures.RATING_DIFFERENCE_PROJECTED,
                                                             RatingEstimatorFeatures.PLAYER_RATING_DIFFERENCE_PROJECTED],
                        pred_column='pred'),
    use_auto_create_performance_calculator=True,
    column_weights=[ColumnWeight(name='points', weight=1)],
)

rating_generator_tuner = UpdateRatingGeneratorTuner(
    team_rating_n_trials=30,
    team_rating_search_ranges=get_default_team_rating_search_range(),
)

match_predictor_tuner = PipelineTuner(
    pipeline=match_predictor_factory,
    rating_generator_tuners=rating_generator_tuner,
    scorer=SklearnScorer(pred_column=match_predictor_factory._predictor.pred_column,
                         scorer_function=mean_absolute_error),
)

match_predictor_tuner.tune(df)



import pickle

import pandas as pd
from sklearn.linear_model import LogisticRegression


from player_performance_ratings.scorer.score import SklearnScorer
from sklearn.metrics import log_loss

from player_performance_ratings.cross_validator.cross_validator import MatchKFoldCrossValidator
from player_performance_ratings.pipeline import Pipeline
from player_performance_ratings.predictor import GameTeamPredictor

from player_performance_ratings.tuner.rating_generator_tuner import UpdateRatingGeneratorTuner
from player_performance_ratings.ratings import UpdateRatingGenerator, ColumnWeight

from player_performance_ratings.data_structures import ColumnNames

from player_performance_ratings.tuner import PipelineTuner, PerformancesGeneratorTuner
from player_performance_ratings.tuner.utils import ParameterSearchRange, get_default_team_rating_search_range

column_names = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='performance',
    league='league',
    position='position',
)
df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])
df['champion_position'] = df['champion'] + df['position']
df['__target'] = df['result']

df = df.drop_duplicates(subset=['gameid', 'teamname', 'playername'])

df = (
    df.assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)

rating_generator = UpdateRatingGenerator(column_names=column_names)

weighter_search_range = {
    column_names.performance:
        [
            ParameterSearchRange(
                name='damagetochampions',
                type='uniform',
                low=0,
                high=0.45
            ),
            ParameterSearchRange(
                name='deaths',
                type='uniform',
                low=0,
                high=.3,
            ),
            ParameterSearchRange(
                name='kills',
                type='uniform',
                low=0,
                high=0.3
            ),
            ParameterSearchRange(
                name='result',
                type='uniform',
                low=0.25,
                high=0.85
            ),
        ]
}

start_rating_search_range = [
    ParameterSearchRange(
        name='team_weight',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='uniform',
        low=20,
        high=100,
    )
]

performance_generator_tuner = PerformancesGeneratorTuner(performances_weight_search_ranges=weighter_search_range,
                                                         lower_is_better_features=["deaths"],
                                                         n_trials=1
                                                         )

rating_generator_tuner = UpdateRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
    start_rating_search_ranges=start_rating_search_range,
    optimize_league_ratings=True,
    team_rating_n_trials=1
)

estimator = LogisticRegression()
predictor = GameTeamPredictor(
    estimator=estimator,
    game_id_colum="gameid",
    team_id_column="teamname",
)
column_weights = [ColumnWeight(name='kills', weight=1),
                  ColumnWeight(name='deaths', weight=1),
                  ColumnWeight(name='damagetochampions', weight=1),
                  ColumnWeight(name='result', weight=1)]

pipeline = Pipeline(
    rating_generators=rating_generator,
    column_weights=column_weights,
    predictor=predictor,
)

cross_validator = MatchKFoldCrossValidator(
    scorer=SklearnScorer(pred_column=pipeline.predictor.pred_column, scorer_function=log_loss),
    match_id_column_name=column_names.match_id,
    n_splits=1,
    date_column_name=column_names.start_date
)

tuner = PipelineTuner(
    performances_generator_tuners=performance_generator_tuner,
    rating_generator_tuners=rating_generator_tuner,
  #  predictor_tuner=predictor_tuner,
    fit_best=True,
    pipeline=pipeline,
    cross_validator=cross_validator,
)
best_match_predictor = tuner.tune(df=df)
pickle.dump(best_match_predictor, open("models/lol_match_predictor", 'wb'))



import math
import time
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.data_structures import PreMatchPlayerRating, PreMatchTeamRating

MATCH_CONTRIBUTION_TO_SUM_VALUE = 1
MODIFIED_RATING_CHANGE_CONSTANT = 1
CERTAIN_SUM = 'certain_sum'


def sigmoid_subtract_half_and_multiply2(value: float, x: float) -> float:
    return (1 / (1 + math.exp(-value / x)) - 0.5) * 2


class PerformancePredictor(ABC):

    @abstractmethod
    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating
                            ) -> float:
        pass


class RatingMeanCustomOpponentRatingPerformancePredictor(PerformancePredictor):

    def __init__(self,
                 opponent_rating_column_name: str,
                 opponent_rating_std: float,
                 opponent_rating_mean: float,
                 player_rating_expected_mean: float = 1000,
                 min_count: int = 2500,
                 coef: float = 0.001757,
                 ):
        self.opponent_rating_column_name = opponent_rating_column_name
        self.opponent_rating_std = opponent_rating_std
        self.opponent_rating_mean = opponent_rating_mean
        self.player_rating_expected_mean = player_rating_expected_mean
        self.min_count = min_count
        self.coef = coef
        self._player_ratings = []
        self._backup_performance_predictor = RatingNonOpponentPerformancePredictor()
        self._sum_rating = 0
        self._rating_count = 0
        self._actual_player_rating_std = 0
        self._actual_player_rating_mean = 0

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating
                            ) -> float:

        opponent_team_rating_value = sum(
            [opponent_player.rating_value for opponent_player in opponent_team_rating.players]) / len(
            opponent_team_rating.players)

        if len(self._player_ratings) == 0:
            actual_player_rating_mean = 0
            weight_actual_data = 0
        else:
            weight_actual_data = min(1, len(self._player_ratings) / 4000)


            if len(self._player_ratings) % 200 == 0 and len(
                    self._player_ratings) >= self.min_count or self._actual_player_rating_std == 0 and len(
                    self._player_ratings) >= self.min_count:
                self._actual_player_rating_std = np.array(self._player_ratings).std()
                self._actual_player_rating_mean = sum(self._player_ratings) / len(self._player_ratings)

        player_rating_mean = (
                                     1 - weight_actual_data) * self.player_rating_expected_mean + weight_actual_data * self._actual_player_rating_mean

        player_rating_std = (
                                    1 - weight_actual_data) * self.opponent_rating_std + weight_actual_data * self._actual_player_rating_std
        opponent_rating_adjusted = (
                                           opponent_team_rating_value - self.opponent_rating_mean + player_rating_mean) / self.opponent_rating_std * player_rating_std

        if len(self._player_ratings) < self.min_count:
            predicted_performance = self._backup_performance_predictor.predict_performance(player_rating=player_rating,
                                                                                           opponent_team_rating=opponent_team_rating,
                                                                                           team_rating=team_rating)


        else:
            historical_average_rating = self._sum_rating / self._rating_count
            net_mean_rating_over_historical_average = player_rating.rating_value * 0.5 + opponent_rating_adjusted * 0.5 - historical_average_rating
            value = self.coef * net_mean_rating_over_historical_average
            predicted_performance = (math.exp(value)) / (1 + math.exp(value))

        self._sum_rating += player_rating.rating_value * 0.5 + opponent_rating_adjusted * 0.5
        self._rating_count += 1

        self._player_ratings.append(player_rating.rating_value)
       # print(predicted_performance, self._backup_performance_predictor.predict_performance(player_rating=player_rating,
     #                                                                                      opponent_team_rating=opponent_team_rating,
      #                                                                                     team_rating=team_rating))
        return predicted_performance


class RatingNonOpponentPerformancePredictor(PerformancePredictor):

    def __init__(self,
                 coef: float = 0.0015,
                 last_sample_count: int = 1500,
                 min_count_for_historical_average: int = 200,
                 historical_average_value_default: float = 1000,
                 ):
        self.coef = coef
        self.last_sample_count = last_sample_count
        self.min_count_for_historical_average = min_count_for_historical_average
        self.historical_average_value_default = historical_average_value_default
        if self.min_count_for_historical_average < 1:
            raise ValueError("min_count_for_historical_average must be positive")
        self._prev_entries_ratings = []

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating) -> float:
        start_index = max(0, len(self._prev_entries_ratings) - self.last_sample_count)
        recent_prev_entries_ratings = self._prev_entries_ratings[start_index:]
        if len(recent_prev_entries_ratings) > self.min_count_for_historical_average:
            historical_average_rating = sum(recent_prev_entries_ratings) / len(recent_prev_entries_ratings)
        else:
            historical_average_rating = self.historical_average_value_default
        net_mean_rating_over_historical_average = player_rating.rating_value - historical_average_rating

        value = self.coef * net_mean_rating_over_historical_average
        prediction = (math.exp(value)) / (1 + math.exp(value))
        self._prev_entries_ratings.append(player_rating.rating_value)

        return prediction


class RatingDifferencePerformancePredictor(PerformancePredictor):

    # TODO: Performance prediction based on team-players sharing time with.
    def __init__(self,
                 rating_diff_coef: float = 0.005757,
                 rating_diff_team_from_entity_coef: float = 0.0,
                 team_rating_diff_coef: float = 0.0,
                 max_predict_value: float = 1,
                 participation_weight_coef: Optional[float] = None,
                 ):
        self.rating_diff_coef = rating_diff_coef
        self.rating_diff_team_from_entity_coef = rating_diff_team_from_entity_coef
        self.team_rating_diff_coef = team_rating_diff_coef
        self.max_predict_value = max_predict_value
        self.participation_weight_coef = participation_weight_coef

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating
                            ) -> float:

        if player_rating.match_performance.team_players_playing_time and isinstance(
                player_rating.match_performance.team_players_playing_time, dict):
            weight_team_rating = 0
            sum_playing_time = 0
            for team_player in team_rating.players:
                if str(team_player.id) in player_rating.match_performance.team_players_playing_time or str(
                        team_player.id) in player_rating.match_performance.team_players_playing_time:
                    weight_team_rating += team_player.rating_value * \
                                          player_rating.match_performance.team_players_playing_time[str(team_player.id)]
                    sum_playing_time += player_rating.match_performance.team_players_playing_time[str(team_player.id)]

            if sum_playing_time > 0:
                team_rating_value = weight_team_rating / sum_playing_time
            else:
                team_rating_value = team_rating.rating_value

        else:
            team_rating_value = team_rating.rating_value

        if player_rating.match_performance.opponent_players_playing_time and isinstance(
                player_rating.match_performance.team_players_playing_time, dict):
            weight_opp_rating = 0
            sum_playing_time = 0
            for opp_player in opponent_team_rating.players:
                if str(opp_player.id) in player_rating.match_performance.opponent_players_playing_time:
                    weight_opp_rating += opp_player.rating_value * \
                                         player_rating.match_performance.opponent_players_playing_time[
                                             str(opp_player.id)]
                    sum_playing_time += player_rating.match_performance.opponent_players_playing_time[
                        str(opp_player.id)]
            if sum_playing_time > 0:
                opp_rating_value = weight_opp_rating / sum_playing_time
            else:
                opp_rating_value = opponent_team_rating.rating_value

        else:
            opp_rating_value = opponent_team_rating.rating_value

        rating_difference = player_rating.rating_value - opp_rating_value

        rating_diff_team_from_entity = team_rating_value - player_rating.rating_value
        team_rating_diff = team_rating_value - opp_rating_value

        value = self.rating_diff_coef * rating_difference + \
                self.rating_diff_team_from_entity_coef * rating_diff_team_from_entity + team_rating_diff * self.team_rating_diff_coef

        prediction = (math.exp(value)) / (1 + math.exp(value))

        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return (1 - self.max_predict_value)

        return prediction


class RatingMeanPerformancePredictor(PerformancePredictor):

    def __init__(self,
                 coef: float = 0.005757,
                 max_predict_value: float = 1,
                 last_sample_count: int = 1500
                 ):
        self.coef = coef
        self.max_predict_value = max_predict_value
        self.last_sample_count = last_sample_count
        self.sum_ratings = []
        self.sum_rating = 0
        self.rating_count = 0

    def predict_performance(self,
                            player_rating: PreMatchPlayerRating,
                            opponent_team_rating: PreMatchTeamRating,
                            team_rating: PreMatchTeamRating) -> float:

        self.sum_ratings.append(player_rating.rating_value)
        self.rating_count += 1
        self.sum_rating += player_rating.rating_value
        start_index = max(0, len(self.sum_ratings) - self.last_sample_count)
        self.sum_ratings = self.sum_ratings[start_index:]
        #  average_rating = sum(self.sum_ratings) / len(self.sum_ratings)
        historical_average_rating = self.sum_rating / self.rating_count
        net_mean_rating_over_historical_average = player_rating.rating_value * 0.5 + opponent_team_rating.rating_value * 0.5 - historical_average_rating

        value = self.coef * net_mean_rating_over_historical_average
        prediction = (math.exp(value)) / (1 + math.exp(value))
        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return (1 - self.max_predict_value)
        return prediction



from abc import ABC, abstractmethod
from typing import Optional, Match

import pandas as pd



class BaseTuner(ABC):

    @abstractmethod
    def tune(self, df: pd.DataFrame, matches: Optional[list[Match]] = None) -> dict[str, float]:
        pass





import pandas as pd

from player_performance_ratings import ColumnNames


def validate_sorting(df: pd.DataFrame, column_names: ColumnNames):
    df_sorted = df.sort_values(
        by=[column_names.start_date, column_names.match_id,
            column_names.team_id, column_names.player_id])

    if not df.equals(df_sorted):
        raise ValueError("df needs to be sorted by date, game_id, team_id, player_id in ascending order")


from .performance_predictor import RatingDifferencePerformancePredictor, RatingMeanPerformancePredictor
from .start_rating_generator import StartRatingGenerator
from .match_rating_generator import MatchRatingGenerator



import pandas as pd


from player_performance_ratings.predictor._base import PredictorTransformer


class ConvertDataFrameToCategoricalTransformer(PredictorTransformer):

    def __init__(self, features: list[str]):
        super().__init__(features=features)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            df = df.assign(**{feature: df[feature].astype('category')})
        return df

    @property
    def features_out(self) -> list[str]:
        return self.features


class SkLearnTransformerWrapper(PredictorTransformer):

    def __init__(self, transformer, features: list[str]):
        self.transformer = transformer
        super().__init__(features=features)
        self._features_out = []

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        try:
            transformed_values = self.transformer.fit_transform(df[self.features]).toarray()
        except AttributeError:
            transformed_values = self.transformer.fit_transform(df[self.features])
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()

        self._features_out = self.transformer.get_feature_names_out().tolist()

        return df.assign(
            **{self._features_out[idx]: transformed_values[:, idx] for idx in range(len(self._features_out))})

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        try:
            transformed_values = self.transformer.transform(df[self.features]).toarray()
        except AttributeError:
            transformed_values = self.transformer.transform(df[self.features])
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()
        return df.assign(
            **{self._features_out[idx]: transformed_values[:, idx] for idx in range(len(self._features_out))})

    @property
    def features_out(self) -> list[str]:
        return self._features_out



import pandas as pd
from lightgbm import LGBMClassifier


from player_performance_ratings.ratings import ColumnWeight

from player_performance_ratings import ColumnNames, PredictColumnNames, PipelineFactory
from player_performance_ratings.ratings.rating_calculators import OpponentAdjustedRatingGenerator

df = pd.read_pickle(r"data/game_player_subsample.pickle")

df = (
    df.assign(team_count=df.groupby("game_id")["team_id"].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)

df = df.sort_values(["start_date", "game_id", "team_id", "player_id"])
df["max_minutes"] = df.groupby(["game_id", "team_id"])["game_minutes"].transform('max')
df['participation_weight'] = df['minutes'] / df['max_minutes']
df.drop(columns=['max_minutes'], inplace=True)

historical_df = df[df['start_date'] < '2023-02-01']
future_df = df[df['start_date'] >= '2023-02-01'][['game_id', 'team_id', 'player_id', 'start_date', 'participation_weight', "location"]]

column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_id",
    performance="plus_minus",
    participation_weight="participation_weight",
)

match_predictor = PipelineFactory(
    rating_generators=[OpponentAdjustedRatingGenerator(column_names=column_names)],
    use_auto_create_performance_calculator=True,
    column_weights=[ColumnWeight(name='plus_minus', weight=1)],
    estimator=LGBMClassifier(),
    group_predictor_by_game_team = True,
    other_categorical_features=["location"]
)

historical_df[PredictColumnNames.TARGET] = historical_df['won']
match_predictor.generate_historical(historical_df)

predicted_future_df = match_predictor.predict(future_df)



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd


class BaseTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass

class BasePostTransformer(ABC):

    def __init__(self, features: list[str], are_estimator_features:bool =True):
        self.features = features
        self._are_estimator_features = are_estimator_features
        self._features_out = []

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    @property
    def estimator_features_out(self) -> list[str]:
        if self._are_estimator_features:
            return self.features_out
        return []


class DifferentGranularityTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, diff_granularity_df: pd.DataFrame, game_player_df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, diff_granularity_df: pd.DataFrame, game_player_df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


